/* -*- c++ -*-
   $Id: controller.cpp,v 1.1 2004/03/17 15:32:25 admin Exp $
*/
#include "controller.h"
#include <ext/stdio_filebuf.h>
#include <iostream>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <boost/lexical_cast.hpp>
#include <boost/lambda/lambda.hpp>
#include <boost/lambda/bind.hpp>
#include <boost/scoped_ptr.hpp>
using namespace std;
using namespace boost;
using namespace boost::lambda;
using namespace shioribox::ctrl;

static void str_setenv(
    const std::string& name,
    const std::string& value,
    bool overwrite = 1) {
    setenv(name.c_str(), value.c_str(), overwrite);

    // DEBUG:
    cerr << "setenv: [" << name << "] -> [" << value << "]" << endl;
}


template<typename _CharT, typename _Traits, typename _Alloc>
static std::basic_istream<_CharT,_Traits>&
getline(
    std::basic_istream<_CharT, _Traits>& __is,
    std::basic_string<_CharT, _Traits, _Alloc>& __str,
    bool accept_lf, bool accept_crlf) {
    __str.clear();
    while (true) {
	_CharT c;
	__is.get(c);
	if (!__is.good()) {
	    break;
	}
	if (c == '\x0d') {
	    // CRLFʤ鼡ʸ⸫롣
	    // Ǥʤ̵롣
	    if (accept_crlf && __is.good()) {
		_CharT next;
		__is.get(next);
		if (next == '\x0a') {
		    // CRLFǽꡣ
		    break;
		}
		else {
		    // ñȤCRϹ˴ޤ롣
		    __str += '\x0d';
		    __str += next;
		}
	    }
	    else {
		__str += '\x0d';
	    }
	}
	else if (c == '\x0a') {
	    // LFʤ顢ǽλ
	    // Ǥʤ̵롣
	    if (accept_lf) {
		break;
	    }
	    else {
		__str += '\x0a';
	    }
	}
	else {
	    __str += c;
	}
    }
    return __is;
}

class controller::impl : noncopyable {
private:
    pid_t pid; // shioriboxץpid
    std::istream input; // ɤ߹ѥȥ꡼
    std::ostream output; // 񤭹ѥȥ꡼
    scoped_ptr<__gnu_cxx::stdio_filebuf<char> > istream_buf; // 
    scoped_ptr<__gnu_cxx::stdio_filebuf<char> > ostream_buf;
    impl()
	:input(NULL), output(NULL) {}
    
    friend class controller;
};

controller::controller(
    const std::string& shioribox_path, // shioriboxХʥؤΥѥ
    const std::string& shiori_name, // shioriboxΥץ̾˻Ȥ̾
    const std::string& shiori_module, // ΤؤΥѥ
    const std::map<std::string, std::string>& additional_env, // ɲäĶѿ
    const std::string& ghost_path) // ٤load()Ϥѥ
    throw (shioribox::ctrl::start_failed_exception) {
    pimpl.reset(new impl());
    
    // ѥפܺ롣
    const int read = 0;
    const int write = 1;
    int pipe_p2c[2], pipe_c2p[2];

    if (pipe(pipe_p2c) != 0) {
	throw start_failed_exception(
	    "Failed to make pipe for controller -> box.");
    }
    
    if (pipe(pipe_c2p) != 0) {
	close(pipe_p2c[write]);
	close(pipe_p2c[read]);
	throw start_failed_exception(
	    "Failed to make pipr for box -> controller.");
    }

    pimpl->pid = fork();
    if (pimpl->pid < 0) {
	// forkԡ
	close(pipe_c2p[write]);
	close(pipe_c2p[read]);
	close(pipe_p2c[write]);
	close(pipe_p2c[read]);
	throw start_failed_exception("Failed to fork(2).");
    }
    else if (pimpl->pid == 0) {
	// ҥץ
	//  controller -> box Υѥפ˽񤫤ʤ
	// box -> controller ΥѥפɤޤʤΤǡĤfdĤ롣
	close(pipe_p2c[write]);
	close(pipe_c2p[read]);

	// ƼĶѿꤹ롣
	str_setenv("SHIORIBOX_MODULE", shiori_module);
	str_setenv("SHIORIBOX_NAME", shiori_name);
	if (ghost_path[ghost_path.size() - 1] != '/') {
	    str_setenv("SHIORIBOX_ROOT", ghost_path+'/');
	}
	else {
	    str_setenv("SHIORIBOX_ROOT", ghost_path);
	}
	str_setenv("SHIORIBOX_FD_IN", lexical_cast<string>(pipe_p2c[read]));
	str_setenv("SHIORIBOX_FD_OUT", lexical_cast<string>(pipe_c2p[write]));
	for_each(
	    additional_env.begin(), additional_env.end(),
	    bind(
		&str_setenv,
		bind(&map<string, string>::value_type::first,  _1),
		bind(&map<string, string>::value_type::second, _1), true));

	// exec¹
	execl(shioribox_path.c_str(), "[shioribox]", NULL);
	//execl(shioribox_path.c_str(), "[shioribox]", "--keep-stderr", NULL);

	// ƤȤϡexec˼ԤƤ롣
	exit(1);
    }
    else {
	// ƥץ
	// controller -> boxΥѥפɤޤʤ
	// box -> controllerΥѥפ˽񤫤ʤΤǡĤfdĤ롣
	close(pipe_p2c[read]);
	close(pipe_c2p[write]);

	// ĤäĤfd饹ȥ꡼롣
	pimpl->istream_buf.reset(
	    new __gnu_cxx::stdio_filebuf<char>(
		pipe_c2p[read], ios_base::in, true, 1));
	pimpl->ostream_buf.reset(
	    new __gnu_cxx::stdio_filebuf<char>(
		pipe_p2c[write], ios_base::out, true, BUFSIZ));
	
	pimpl->input.rdbuf(pimpl->istream_buf.get());
	pimpl->output.rdbuf(pimpl->ostream_buf.get());

	cerr <<
	    "shioribox controller: subprocess [" << shiori_name <<
	    "] started on pid " << pimpl->pid << endl;
    }
}

controller::~controller() {
    // ܤΥѥפĤ롣
    pimpl->input.rdbuf(NULL);
    pimpl->output.rdbuf(NULL);
    pimpl->istream_buf.reset();
    pimpl->ostream_buf.reset();

    // shioriboxϽλˤʤäƤΤǡλԤġ
    int status;
    waitpid(pimpl->pid, &status, 0);

    cerr <<
	"shioribox controller: subprocess exitted with status " <<
	WEXITSTATUS(status) << endl;
}

std::string controller::request(const std::string& req)
    throw (out_of_control_exception) {
    
    // shioribox˥ꥯƤ򤽤Τޤ롣
    pimpl->output << req << flush;

    // ֤äƤƤ򤽤Τޤ֤ԤϽü
    string reply;
    string linebuf;
    const string CRLF = "\x0d\x0a";
    while (pimpl->input.good()) {
	getline(pimpl->input, linebuf, false, true);
	if (linebuf.length() == 0) {
	    // ԡreply˥ǡίޤäƤСǽλ
	    if (reply.length() > 0) {
		reply += CRLF;
		return reply;
	    }
	}
	else {
	    reply += linebuf;
	    reply += CRLF;
	}
    }

    // 褿Ȥϡ꽪ˡѥפĤƤ
    // ¿ʬҥץϤ⤦ǤΤǡʥ0äƤߤ롣
    if (kill(pimpl->pid, 0) == 0) {
	// ҤƤ뤬
	throw out_of_control_exception(
	    string() + "Child process (pid " + lexical_cast<string>(pimpl->pid) +
	    ") is still alive, but it closed the pipe.");
    }
    else {
	throw out_of_control_exception(
	    "Child process died suddenly.");
    }
}
