/* -*- c++ -*-
   $Id: shioribox_controller_jni.cpp,v 1.1 2004/03/17 15:32:25 admin Exp $
*/
#include "shioribox_controller_jni.h"
#include "controller.h"
#include <vector>
#include <boost/lexical_cast.hpp>
#include <boost/scoped_ptr.hpp>
using namespace std;
using namespace boost;
using namespace shioribox::ctrl;

class controller_jni::impl : noncopyable {
private:
    scoped_ptr<controller> ctrl;
    
    static vector<shared_ptr<controller_jni> > contexts;
    static bool validate_context_id(JNIEnv *env, int context_id);
    friend class controller_jni;
};
vector<shared_ptr<controller_jni> > controller_jni::impl::contexts;

static string make_utf8_string_from_jstring(JNIEnv *env, jstring jstr) {
    const char* chars = env->GetStringUTFChars(jstr, NULL);
    string str(chars, env->GetStringUTFLength(jstr));
    env->ReleaseStringUTFChars(jstr, chars);
    return str;
}

static string make_string_from_jbyteArray(JNIEnv *env, jbyteArray jbytes) {
    char *bytes = reinterpret_cast<char*>(
	env->GetByteArrayElements(jbytes, NULL));
    long len = env->GetArrayLength(jbytes);
    string str(bytes, len);
    env->ReleaseByteArrayElements(
	jbytes, reinterpret_cast<jbyte*>(bytes), JNI_ABORT);
    return str;
}

static jbyteArray make_jbyteArray_from_string(JNIEnv *env, const string& str) {
    long len = str.length();
    jbyteArray jbytes = env->NewByteArray(len);
    env->SetByteArrayRegion(
	jbytes, 0, len, const_cast<jbyte*>(
	    reinterpret_cast<const jbyte*>(str.c_str())));
    return jbytes;
}

static void throw_new_java_exception(
    JNIEnv *env,
    const std::string& exception_class_name,
    const std::string& message) {
    jclass c_exception = env->FindClass(exception_class_name.c_str());
    if (c_exception) {
	env->ThrowNew(c_exception, message.c_str());
    }
}

controller_jni::controller_jni()
    :pimpl(new impl()) {}

int controller_jni::open(
    JNIEnv *env,
    const jstring shioribox_path, // shioriboxХʥؤΥѥ
    const jstring shiori_name, // shioriboxΥץ̾˻Ȥ̾
    const jstring shiori_module, // ΤؤΥѥ
    const jobject additional_env, // ɲäĶѿnullġHashtable<String, String>
    const jstring ghost_path) { // ٤load()Ϥѥ

#define CHECK(x) if(!(x)) return -1

    // additional_envmap<string, string>롣
    map<string, string> map_additional_env;
    if (additional_env != NULL) {
	jclass c_hashtable = env->FindClass("java/util/Hashtable");
	CHECK(c_hashtable);
	jmethodID m_hashtable_entrySet =
	    env->GetMethodID(c_hashtable, "entrySet", "()Ljava/util/Set;");
	CHECK(m_hashtable_entrySet);
	jobject o_set = env->CallObjectMethod(additional_env, m_hashtable_entrySet);
	CHECK(o_set);

	jclass c_set = env->FindClass("java/util/Set");
	CHECK(c_set);
	jmethodID m_set_iterator =
	    env->GetMethodID(c_set, "iterator", "()Ljava/util/Iterator;");
	CHECK(m_set_iterator);
	jobject o_iterator = env->CallObjectMethod(o_set, m_set_iterator);
	CHECK(o_iterator);

	jclass c_iterator = env->FindClass("java/util/Iterator");
	CHECK(c_iterator);
	jmethodID m_iterator_hasNext =
	    env->GetMethodID(c_iterator, "hasNext", "()Z");
	CHECK(m_iterator_hasNext);
	jmethodID m_iterator_next =
	    env->GetMethodID(c_iterator, "next", "()Ljava/lang/Object;");
	CHECK(m_iterator_next);
	
	jclass c_map_entry = env->FindClass("java/util/Map$Entry");
	CHECK(c_map_entry);
	jmethodID m_entry_getKey =
	    env->GetMethodID(c_map_entry, "getKey", "()Ljava/lang/Object;");
	CHECK(m_entry_getKey);
	jmethodID m_entry_getValue =
	    env->GetMethodID(c_map_entry, "getValue", "()Ljava/lang/Object;");
	CHECK(m_entry_getValue);

	jclass c_string = env->FindClass("java/lang/String");
	CHECK(c_string);
	
	while (1) {
	    if (!env->CallBooleanMethod(o_iterator, m_iterator_hasNext)) {
		break;
	    }
	    jobject o_entry = env->CallObjectMethod(o_iterator, m_iterator_next);
	    CHECK(o_entry);
	    
	    jobject o_key = env->CallObjectMethod(o_entry, m_entry_getKey);
	    jobject o_value = env->CallObjectMethod(o_entry, m_entry_getValue);
	    if (!o_key || !env->IsInstanceOf(o_key, c_string) ||
		!o_value || !env->IsInstanceOf(o_value, c_string)) {
		return -1;
	    }

	    // o_keyo_valueʸǤϳΤ줿Τǡ
	    // UTF-8ʸФmap롣
	    map_additional_env[
		make_utf8_string_from_jstring(env, static_cast<jstring>(o_key))] =
		make_utf8_string_from_jstring(env, static_cast<jstring>(o_value));
	}
    }
    
    // Ƥcontext idõ
    int context_id = -1;
    for (int i = 0; i < impl::contexts.size(); i++) {
	if (impl::contexts[i].get() == NULL) {
	    context_id = i;
	    break;
	}
    }
    if (context_id == -1) {
	// ̵Τǿɲá
	impl::contexts.push_back(shared_ptr<controller_jni>());
	context_id = impl::contexts.size() - 1;
    }

    impl::contexts[context_id].reset(new controller_jni());
    try {
	controller* ctrl = new controller(
	    make_utf8_string_from_jstring(env, shioribox_path),
	    make_utf8_string_from_jstring(env, shiori_name),
	    make_utf8_string_from_jstring(env, shiori_module),
	    map_additional_env,
	    make_utf8_string_from_jstring(env, ghost_path));
	impl::contexts[context_id]->pimpl->ctrl.reset(ctrl);
    }
    catch (start_failed_exception& e) {
	// open()̵äˤRuntimeExceptionꤲ롣
	throw_new_java_exception(
	    env, "java/lang/RuntimeException", e.what());
	return -1;
    }
    return context_id;
}

bool controller_jni::impl::validate_context_id(JNIEnv *env, int context_id) {
    if (context_id >= 0 &&
	context_id < contexts.size() &&
	contexts[context_id].get() != NULL) {
	return true;
    }
    else {
	string msg = "Invalid ContextID: " + lexical_cast<string>(context_id);
	throw_new_java_exception(
	    env, "java/lang/RuntimeException", msg);
	return false;
    }
}

jbyteArray controller_jni::request(
    JNIEnv *env,
    int context_id,
    const jbyteArray req) {
    // context_id
    if (!impl::validate_context_id(env, context_id)) {
	return NULL;
    }
    
    controller_jni& context = *(impl::contexts[context_id]);
    
    string req_str = make_string_from_jbyteArray(env, req);
    string rep_str;
    try {
	rep_str = context.pimpl->ctrl->request(req_str);
    }
    catch (out_of_control_exception& e) {
	throw_new_java_exception(
	    env, "java/lang/RuntimeException", e.what());
	return NULL;
    }
    return make_jbyteArray_from_string(env, rep_str);
}

void controller_jni::close(
    JNIEnv *env,
    int context_id) {
    // context_id
    if (!impl::validate_context_id(env, context_id)) {
	return;
    }
    impl::contexts[context_id].reset();

    // κǸˤ롢contextƺ
    while (
	impl::contexts.size() > 0 &&
	impl::contexts[impl::contexts.size()-1].get() == NULL) {
	
	impl::contexts.pop_back();
    }
}
