/* -*- c++ -*-
   $Id: shiori_session.cpp,v 1.2 2004/03/17 15:32:25 admin Exp $
*/
#include "shiori_session.h"
#include <exception>
#include <vector>
#include <sstream>
#include <boost/format.hpp>
using namespace std;
using namespace shioribox;
using namespace boost;

template<typename _CharT, typename _Traits, typename _Alloc>
std::basic_istream<_CharT,_Traits>&
getline(
    std::basic_istream<_CharT, _Traits>& __is,
    std::basic_string<_CharT, _Traits, _Alloc>& __str,
    bool accept_lf, bool accept_crlf) {
    __str.clear();
    while (true) {
	_CharT c;
	__is.get(c);
	if (!__is.good()) {
	    break;
	}
	if (c == '\x0d') {
	    // CRLFʤ鼡ʸ⸫롣
	    // Ǥʤ̵롣
	    if (accept_crlf && __is.good()) {
		_CharT next;
		__is.get(next);
		if (next == '\x0a') {
		    // CRLFǽꡣ
		    break;
		}
		else {
		    // ñȤCRϹ˴ޤ롣
		    __str += '\x0d';
		    __str += next;
		}
	    }
	    else {
		__str += '\x0d';
	    }
	}
	else if (c == '\x0a') {
	    // LFʤ顢ǽλ
	    // Ǥʤ̵롣
	    if (accept_lf) {
		break;
	    }
	    else {
		__str += '\x0a';
	    }
	}
	else {
	    __str += c;
	}
    }
    return __is;
}

template<typename _CharT, typename _Traits, typename _Alloc>
std::basic_string<_CharT, _Traits, _Alloc>&
replace(
    std::basic_string<_CharT, _Traits, _Alloc>& __str,
    const std::basic_string<_CharT, _Traits, _Alloc>& __from,
    const std::basic_string<_CharT, _Traits, _Alloc>& __to) {
    
    typedef std::basic_string<_CharT, _Traits, _Alloc> _strtype;

    typename _strtype::size_type pos = 0;
    while (1) {
	typename _strtype::size_type found_pos = __str.find(__from, pos);
	if (found_pos == _strtype::npos) {
	    // ⤦̵
	    break;
	}
	__str.replace(
	    __str.begin() + found_pos,
	    __str.begin() + found_pos + __from.length(),
	    __to);
	pos = found_pos + __to.length();
    }
    
    return __str;
}

int shiori_session::start(const std::string& rootdir) {
    if (started_already) {
	throw bad_exception(); // start()ϰ٤Ƥ٤ʤ
    }
    started_already = true;

    if (!shiori.load(rootdir)) {
	// loadԡ
	// stderrĤƤ뤫Τʤ񤯤񤤤Ƥߤ롣
	cerr << "load() failed" << endl;
	return 1;
    }

    // 롼׳ϡ
    // ɤ٤˥ХåեίƹԤԤ褿
    // ꥯȤνʤΤǡrequest()롣
    // inputĤ줿롼׽λ
    string reqbuf;
    string linebuf;
    const string CRLF = "\x0d\x0a";
    const string LF = "\x0a";
    while (input.good()) {
	// CRLFѴ⡼ɤǤϡԥLFCRLF졢
	// ɤCRLF줹롣
	// ǤʤCRLFΤߤ롣
	getline(input, linebuf, convert_crlf, true);
	if (linebuf.length() == 0) {
	    // ԡreqbuf˥ǡίޤäƤХꥯȯԡ
	    if (reqbuf.length() > 0) {
		reqbuf += CRLF;

		// reqbufн
		fix_dialogue(reqbuf, true);
		
		string reply = shiori.request(reqbuf);
		reqbuf.clear();

		// replyн
		fix_dialogue(reply, false);

		// CRLFѴ⡼ɤǤϡƤCRLFLFѴ롣
		if (convert_crlf) {
		    replace(reply, CRLF, LF);
		}

		output << reply << flush;
	    }
	}
	else {
	    // ԤǤϤʤ
	    reqbuf += linebuf;
	    reqbuf += CRLF;
	}
    }

    if (!shiori.unload()) {
	// unloadԡ
	// stderrĤƤ뤫Τʤ񤯤񤤤Ƥߤ롣
	cerr << "unload() failed" << endl;
	return 1;
    }

    return 0;
}

void shiori_session::fix_dialogue(std::string& dialogue, bool is_request) {
    // ޤCRLFʬ䡣
    vector<string> lines;
    
    istringstream is(dialogue);
    string linebuf;
    while (is.good()) {
	getline(is, linebuf, false, true);
	lines.push_back(linebuf);
    }
    // Ǹ;ʬʶԤդΤǾä
    if (lines.size() > 0 && (lines.end()-1)->empty()) {
	lines.erase(lines.end()-1);
    }

    // ȤꤢΤ϶Ԥ
    bool dirty = false;
    for (unsigned i = 0; i < lines.size(); i++) {
	if (lines[i].empty()) {
	    // ԤäɤΤϺǽԤ
	    if (i != lines.size()-1) {
		warn_strange_dialogue(
		    is_request,
		    lines,
		    (i == 0 ?
		     "Blank line is in front of the header. fixed..." :
		     "Blank line is in the middle. fixed..."),
		    i);
		// ζԤϺ롣
		lines.erase(lines.begin() + i);
		i--;
		dirty = true;
	    }
	}
	else {
	    // ǽԤˤ϶Ԥ̵Фʤʤ
	    if (i == lines.size()-1) {
		warn_strange_dialogue(
		    is_request,
		    lines,
		    "Blank line is missing from the footer. fixed...",
		    i);
		// θ˶Ԥɲä롣
		lines.push_back(string());
		dirty = true;
	    }
	}
    }

    if (dirty) {
	// ƹ
	dialogue.clear();
	for (vector<string>::const_iterator ite = lines.begin();
	     ite != lines.end(); ite++) {
	    dialogue += *ite;
	    dialogue += "\x0d\x0a";
	}
    }
}

void shiori_session::warn_strange_dialogue(
    bool is_request,
    const std::vector<std::string>& lines,
    const std::string& message,
    unsigned position) {

    if (is_request) {
	cerr << "# ----- Warning: malformed request has come from the client ----- #" << endl;
    }
    else {
	cerr << "# ----- Warning:  malformed reply has come from the server  ----- #" << endl;
    }

    for (vector<string>::size_type i = 0; i < lines.size(); i++) {
	cerr << format(
	    i == position ?
	    "=> %2d: %s" :
	    "   %2d: %s") % i % lines[i] << endl;
    }

    cerr << message << endl;
}
