/* -*- c++ -*-
  $Id: shioribox.cpp,v 1.2 2004/03/17 15:32:25 admin Exp $
*/
#include "config.h"
#include "shiori_accessor.h"
#include "shiori_session.h"

#include <iostream>
#include <string>
#include <vector>
#include <fstream>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ext/stdio_filebuf.h>
#include <boost/lexical_cast.hpp>
#include <boost/scoped_ptr.hpp>
using namespace std;
using namespace boost;
using namespace shioribox;

static string str_getenv(const string& name) {
    char* var = getenv(name.c_str());
    if (var == NULL) {
	return string();
    }
    else {
	return string(var);
    }
}

static string str_getcwd() {
    char* cwd = getcwd(NULL, 0);
    string str(cwd);
    free(cwd);
    return str;
}

static void show_help() {
    cout <<
	"This is shioribox; executes shiori module in a process." << endl <<
	"Public Domain - feel free to modify/redistribute this program." << endl <<
	endl <<
	"Usage:" << endl <<
	"  shioribox [options..] <module>" << endl <<
	"or" << endl <<
	"  SHIORIBOX_MODULE=<module> shioribox [options..]" << endl <<
	endl <<
	"Options:" << endl <<
	"  -h --help          print this message" << endl <<
	"  -c --crlf          convert input(LF to CRLF) and output(CRLF to LF)" << endl <<
	"  -e --keep-stderr   don't discard any output which shiori prints to stderr" << endl <<
	endl <<
	"Environment Variables:" << endl <<
	"  SHIORIBOX_MODULE   Specify module(library) file path to use." << endl <<
	"                     You can also specify this by argument." << endl <<
	"                     If both are specified, argument's one will be used." << endl <<
	"  SHIORIBOX_NAME     Specify short name of shiori." << endl <<
	"                     When specified, shioribox makes its process name" << endl <<
	"                     like `[shioribox - aya]'." << endl <<
	"                     This feature can be unusable on some platforms." << endl <<
	"  SHIORIBOX_ROOT     Specify a directory to be passed to load() of the module." << endl <<
	"                     If omitted, the current directory will be used." << endl <<
	"  SHIORIBOX_FD_IN    " << endl <<
	"  SHIORIBOX_FD_OUT   In default, shioribox uses stdin(fd 0) and stdout(fd 1)" << endl <<
	"                     as I/O of shiori module, and closes original stdin/out"  << endl <<
	"                     not to corrupt the protocol (so modules will print" << endl <<
	"                     nothing to stdout)." << endl <<
	"                     You can change this behavior by fork(2)ing with these" << endl <<
	"                     two fd:s open (and of course, setting these env:s)." << endl <<
	"                     When specified, shioribox no longer closes original" << endl <<
	"                     stdin/out." << endl <<
	endl <<
	"Please report bugs to <" << PACKAGE_BUGREPORT << ">." << endl;
}

int main(int argc, char* argv[]) {
    vector<string> args;
    for (int i = 0; i < argc; i++) {
	args.push_back(argv[i]);
    }

    // argv[0]
    string env_shioriname = str_getenv("SHIORIBOX_NAME");
    if (env_shioriname.length() > 0) {
	sprintf(argv[0], "[shioribox - %s]", env_shioriname.c_str());
    }
    else {
	sprintf(argv[0], "[shioribox - unknown]");
    }
    
    /*
      ɤ饤֥ꤹˡġ
      ܤϥѥ᡼ȤϤˡǡܤϴĶѿ SHIORIBOX_MODULE ˻ꤹˡ
      ɤȤʤСإץåɽ
      ξȤ줿ѥ᡼ͥ褹롣
    */
    string env_libfile = str_getenv("SHIORIBOX_MODULE");
    bool opt_help = false;
    bool opt_crlf = false;
    bool opt_keep_stderr = false;
    string opt_libfile;
    for (vector<string>::const_iterator ite = args.begin() + 1;
	 ite != args.end(); ite++) {
	if (*ite == "-h" || *ite == "--help") {
	    opt_help = true;
	}
	else if (*ite == "-c" || *ite == "--crlf") {
	    opt_crlf = true;
	}
	else if (*ite == "-e" || *ite == "--keep-stderr") {
	    opt_keep_stderr = true;
	}
	else if ((*ite)[0] == '-') {
	    cerr << "Unrecognized option: " << *ite << endl;
	    return 1;
	}
	else {
	    env_libfile = *ite;
	}
    }
    
    if (opt_help) {
	show_help();
	return 0;
    }

    // ɤ٥⥸塼Ȥ
    string shiori_to_use;
    if (opt_libfile.length() > 0) {
	shiori_to_use = opt_libfile;
    }
    else {
	if (env_libfile.length() > 0) {
	    shiori_to_use = env_libfile;
	}
	else {
	    cerr <<
		"Error:" << endl <<
		"  Specify module(library) file to use by argument or environment variable." << endl <<
		"  See `shioribox -h' for details." << endl;
	    return 0;
	}
    }
    scoped_ptr<shiori_accessor> shiori;
    try {
	shiori.reset(new shiori_accessor(shiori_to_use));
    }
    catch (bad_module_exception& e) {
	cerr << e.what() << endl;
	return 1;
    }

    // 롼ȥǥ쥯ȥ
    string rootdir = str_getenv("SHIORIBOX_ROOT");
    if (rootdir.length() == 0) {
	rootdir = str_getcwd();
    }

    // եǥץϢ
    int fd_in = -1, fd_out = -1;
    try {
	int in = lexical_cast<int>(str_getenv("SHIORIBOX_FD_IN"));
	int out = lexical_cast<int>(str_getenv("SHIORIBOX_FD_OUT"));
	// Υǥץϼºߤ뤫
	struct stat sb;
	bool fdin_valid = (fstat(in, &sb) == 0);
	bool fdout_valid = (fstat(out, &sb) == 0);
	if (fdin_valid && fdout_valid) {
	    fd_in = in;
	    fd_out = out;
	}
	else {
	    // ʤλ
	    cerr << "Error:" << endl << "  Specified pid ";
	    if (!fdin_valid) {
		cerr << in << "(SHIORIBOX_FD_IN)";
	    }
	    if (!fdout_valid) {
		if (!fdin_valid) {
		    cerr << ", ";
		}
		cerr << out << "(SHIORIBOX_FD_OUT)";
	    }
	    if ((fdin_valid ? 1 : 0) + (fdout_valid ? 1 : 0) == 1) {
		cerr << " is invalid." << endl;
	    }
	    else {
		cerr << " are invalid." << endl;
	    }
	    return 1;
	}
    }
    catch (bad_lexical_cast &) {}
    if (fd_in == -1 || fd_out == -1) {
	// stdinstdoutȤΤǡdupƤ鸵fdĤ롣
	fd_in = dup(0); close(0);
	fd_out = dup(1); close(1);
    }
    __gnu_cxx::stdio_filebuf<char> fb_in(fd_in, ios_base::in, false, 1);
    __gnu_cxx::stdio_filebuf<char> fb_out(fd_out, ios_base::out, false, BUFSIZ);
    istream shiori_in(&fb_in);
    ostream shiori_out(&fb_out);

    // ɬפʤstderrĤ塢ShioriIO򳫻ϡ
    if (!opt_keep_stderr) {
	close(2);
    }
    
    shiori_session session(
	opt_crlf, *shiori, shiori_in, shiori_out);
    int retval = session.start(rootdir);

    // ȥ꡼Ȥ˴ˡμ¤Υ
    shiori_in.rdbuf(NULL);
    shiori_out.rdbuf(NULL);

    return retval;
}
